@@ -1,7 +1,9 @@
 import os
 
 import numpy as np
+import resampy
 import torch
+import tqdm
 
 import torchcrepe
 
@@ -15,10 +17,12 @@ __all__ = ['CENTS_PER_BIN',
            'embed',
            'embed_from_file',
            'embed_from_file_to_file',
+           'embed_from_files_to_files',
            'infer',
            'predict',
            'predict_from_file',
            'predict_from_file_to_file',
+           'predict_from_files_to_files',
            'preprocess',
            'postprocess',
            'resample']
@@ -44,12 +48,13 @@ UNVOICED = np.nan
 
 def predict(audio,
             sample_rate,
-            hop_length,
+            hop_length=None,
             fmin=50.,
             fmax=MAX_FMAX,
             model='full',
             decoder=torchcrepe.decode.viterbi,
-            return_harmonicity=False):
+            return_harmonicity=False,
+            batch_size=None):
     """Performs pitch estimation
 
     Arguments
@@ -69,32 +74,53 @@ def predict(audio,
             The decoder to use. See decode.py for decoders.
         return_harmonicity (bool)
             Whether to also return the network confidence
+        batch_size (int)
+            The number of frames per batch
 
     Returns
-        pitch (torch.tensor [shape=(1, time / hop_length)])
-        (Optional) harmonicity(torch.tensor [shape=(1, time / hop_length)])
+        pitch (torch.tensor [shape=(1, 1 + int(time // hop_length))])
+        (Optional) harmonicity (torch.tensor
+                                [shape=(1, 1 + int(time // hop_length))])
     """
-    # Preprocess audio
-    frames = preprocess(audio, sample_rate, hop_length)
+    results = []
 
-    # Infer independent probabilities for each pitch bin
-    probabilities = infer(frames, model)
+    # Postprocessing breaks gradients, so just don't compute them
+    with torch.no_grad():
 
-    # shape=(batch, 360, time / hop_length)
-    probabilities = probabilities.reshape(
-        audio.size(0), -1, PITCH_BINS).transpose(1, 2)
+        # Preprocess audio
+        for frames in preprocess(audio, sample_rate, hop_length, batch_size):
 
-    # Convert probabilities to F0 and harmonicity
-    return postprocess(probabilities, fmin, fmax, decoder, return_harmonicity)
+            # Infer independent probabilities for each pitch bin
+            probabilities = infer(frames, model)
+
+            # shape=(batch, 360, time / hop_length)
+            probabilities = probabilities.reshape(
+                audio.size(0), -1, PITCH_BINS).transpose(1, 2)
+
+            # Convert probabilities to F0 and harmonicity
+            results.append(postprocess(probabilities,
+                                       fmin,
+                                       fmax,
+                                       decoder,
+                                       return_harmonicity))
+
+    # Split pitch and harmonicity
+    if return_harmonicity:
+        pitch, harmonicity = zip(*results)
+        return torch.cat(pitch, 1), torch.cat(harmonicity, 1)
+
+    # Concatenate
+    return torch.cat(results, 1)
 
 
 def predict_from_file(audio_file,
-                      hop_length,
+                      hop_length=None,
                       fmin=50.,
                       fmax=MAX_FMAX,
                       model='full',
                       decoder=torchcrepe.decode.viterbi,
                       return_harmonicity=False,
+                      batch_size=None,
                       device='cpu'):
     """Performs pitch estimation from file on disk
 
@@ -113,12 +139,15 @@ def predict_from_file(audio_file,
             The decoder to use. See decode.py for decoders.
         return_harmonicity (bool)
             Whether to also return the network confidence
+        batch_size (int)
+            The number of frames per batch
         device (string)
             The device used to run inference
 
     Returns
-        pitch (torch.tensor [shape=(1, time / hop_length)])
-        (Optional) harmonicity(torch.tensor [shape=(1, time / hop_length)])
+        pitch (torch.tensor [shape=(1, 1 + int(time // hop_length))])
+        (Optional) harmonicity (torch.tensor
+                                [shape=(1, 1 + int(time // hop_length))])
     """
     # Load audio
     audio, sample_rate = torchcrepe.load.audio(audio_file)
@@ -131,29 +160,31 @@ def predict_from_file(audio_file,
                    fmax,
                    model,
                    decoder,
-                   return_harmonicity)
+                   return_harmonicity,
+                   batch_size)
 
 
 def predict_from_file_to_file(audio_file,
-                              hop_length,
                               output_pitch_file,
                               output_harmonicity_file=None,
+                              hop_length=None,
                               fmin=50.,
                               fmax=MAX_FMAX,
                               model='full',
                               decoder=torchcrepe.decode.viterbi,
+                              batch_size=None,
                               device='cpu'):
     """Performs pitch estimation from file on disk
 
     Arguments
         audio_file (string)
             The file to perform pitch tracking on
-        hop_length (int)
-            The hop_length in samples
         output_pitch_file (string)
             The file to save predicted pitch
         output_harmonicity_file (string or None)
             The file to save predicted harmonicity
+        hop_length (int)
+            The hop_length in samples
         fmin (float)
             The minimum allowable frequency in Hz
         fmax (float)
@@ -162,6 +193,8 @@ def predict_from_file_to_file(audio_file,
             The model capacity. One of 'full' or 'tiny'.
         decoder (function)
             The decoder to use. See decode.py for decoders.
+        batch_size (int)
+            The number of frames per batch
         device (string)
             The device used to run inference
     """
@@ -173,6 +206,7 @@ def predict_from_file_to_file(audio_file,
                                    model,
                                    decoder,
                                    output_harmonicity_file is not None,
+                                   batch_size,
                                    device)
 
     # Save to disk
@@ -183,12 +217,66 @@ def predict_from_file_to_file(audio_file,
         torch.save(prediction.detach(), output_pitch_file)
 
 
+def predict_from_files_to_files(audio_files,
+                                output_pitch_files,
+                                output_harmonicity_files=None,
+                                hop_length=None,
+                                fmin=50.,
+                                fmax=MAX_FMAX,
+                                model='full',
+                                decoder=torchcrepe.decode.viterbi,
+                                batch_size=None,
+                                device='cpu'):
+    """Performs pitch estimation from files on disk without reloading model
+
+    Arguments
+        audio_files (list[string])
+            The files to perform pitch tracking on
+        output_pitch_files (list[string])
+            The files to save predicted pitch
+        output_harmonicity_files (list[string] or None)
+            The files to save predicted harmonicity
+        hop_length (int)
+            The hop_length in samples
+        fmin (float)
+            The minimum allowable frequency in Hz
+        fmax (float)
+            The maximum allowable frequency in Hz
+        model (string)
+            The model capacity. One of 'full' or 'tiny'.
+        decoder (function)
+            The decoder to use. See decode.py for decoders.
+        batch_size (int)
+            The number of frames per batch
+        device (string)
+            The device used to run inference
+    """
+    if output_harmonicity_files is None:
+        output_harmonicity_files = len(audio_files) * [None]
+
+    # Setup iterator
+    iterator = zip(audio_files, output_pitch_files, output_harmonicity_files)
+    iterator = tqdm.tqdm(iterator, dynamic_ncols=True)
+    for audio_file, output_pitch_file, output_harmonicity_file in iterator:
+
+        # Predict a file
+        predict_from_file_to_file(audio_file,
+                                  output_pitch_file,
+                                  output_harmonicity_file,
+                                  hop_length,
+                                  fmin,
+                                  fmax,
+                                  model,
+                                  decoder,
+                                  batch_size,
+                                  device)
+
 ###############################################################################
 # Crepe pitch embedding
 ###############################################################################
 
 
-def embed(audio, sample_rate, hop_length, model='full'):
+def embed(audio, sample_rate, hop_length=None, model='full', batch_size=None):
     """Embeds audio to the output of CREPE's fifth maxpool layer
 
     Arguments
@@ -200,21 +288,33 @@ def embed(audio, sample_rate, hop_length, model='full'):
             The hop_length in samples
         model (string)
             The model capacity. One of 'full' or 'tiny'.
+        batch_size (int)
+            The number of frames per batch
 
     Returns
-        embedding (torch.tensor [shape=(1, time / hop_length, 32, -1)])
+        embedding (torch.tensor [shape=(1,
+                                        1 + int(time // hop_length), 32, -1)])
     """
+    results = []
+
     # Preprocess audio
-    frames = preprocess(audio, sample_rate, hop_length)
+    for frames in preprocess(audio, sample_rate, hop_length, batch_size):
 
-    # Infer pitch embeddings
-    embedding = infer(frames, model, embed=True)
+        # Infer pitch embeddings
+        embedding = infer(frames, model, embed=True)
 
-    # shape=(batch, time / hop_length, 32, embedding_size)
-    return embedding.reshape(audio.size(0), frames.size(0), 32, -1)
+        # shape=(batch, time / hop_length, 32, embedding_size)
+        results.append(embedding.reshape(audio.size(0), frames.size(0), 32, -1))
+
+    # Concatenate
+    return torch.cat(results, 1)
 
 
-def embed_from_file(audio_file, hop_length, model='full', device='cpu'):
+def embed_from_file(audio_file,
+                    hop_length=None,
+                    model='full',
+                    batch_size=None,
+                    device='cpu'):
     """Embeds audio from disk to the output of CREPE's fifth maxpool layer
 
     Arguments
@@ -224,23 +324,27 @@ def embed_from_file(audio_file, hop_length, model='full', device='cpu'):
             The hop_length in samples
         model (string)
             The model capacity. One of 'full' or 'tiny'.
+        batch_size (int)
+            The number of frames per batch
         device (string)
             The device to run inference on
 
     Returns
-        embedding (torch.tensor [shape=(1, time / hop_length, 32, -1)])
+        embedding (torch.tensor [shape=(1,
+                                        1 + int(time // hop_length), 32, -1)])
     """
     # Load audio
     audio, sample_rate = torchcrepe.load.audio(audio_file)
 
     # Embed
-    return embed(audio.to(device), sample_rate, hop_length, model)
+    return embed(audio.to(device), sample_rate, hop_length, model, batch_size)
 
 
 def embed_from_file_to_file(audio_file,
-                            hop_length,
                             output_file,
+                            hop_length=None,
                             model='full',
+                            batch_size=None,
                             device='cpu'):
     """Embeds audio from disk and saves to disk
 
@@ -253,17 +357,59 @@ def embed_from_file_to_file(audio_file,
             The file to save the embedding
         model (string)
             The model capacity. One of 'full' or 'tiny'.
+        batch_size (int)
+            The number of frames per batch
         device (string)
             The device to run inference on
-
-    Returns
-        embedding (torch.tensor [shape=(1, time / hop_length, 32, -1)])
     """
-    # Embed
-    embedding = embed_from_file(audio_file, hop_length, model, device)
+    # No use computing gradients if we're just saving to file
+    with torch.no_grad():
 
-    # Save to disk
-    torch.save(embedding.detach(), output_file)
+        # Embed
+        embedding = embed_from_file(audio_file,
+                                    hop_length,
+                                    model,
+                                    batch_size,
+                                    device)
+
+        # Save to disk
+        torch.save(embedding.detach(), output_file)
+
+
+def embed_from_files_to_files(audio_files,
+                              output_files,
+                              hop_length=None,
+                              model='full',
+                              batch_size=None,
+                              device='cpu'):
+    """Embeds audio from disk and saves to disk without reloading model
+
+    Arguments
+        audio_files (list[string])
+            The wav files containing the audio to embed
+        output_files (list[string])
+            The files to save the embeddings
+        hop_length (int)
+            The hop_length in samples
+        model (string)
+            The model capacity. One of 'full' or 'tiny'.
+        batch_size (int)
+            The number of frames per batch
+        device (string)
+            The device to run inference on
+    """
+    # Setup iterator
+    iterator = zip(audio_files, output_files)
+    iterator = tqdm.tqdm(iterator, dynamic_ncols=True)
+    for audio_file, output_file in iterator:
+
+        # Embed a file
+        embed_from_file_to_file(audio_file,
+                                output_file,
+                                hop_length,
+                                model,
+                                batch_size,
+                                device)
 
 
 ###############################################################################
@@ -283,21 +429,14 @@ def infer(frames, model='full', embed=False):
             Whether to stop inference at the intermediate embedding layer
 
     Returns
-        logits (torch.tensor [shape=(time / hop_length, 360)]) OR
-        embedding (torch.tensor [shape=(time / hop_length, embedding_size)])
+        logits (torch.tensor [shape=(1 + int(time // hop_length), 360)]) OR
+        embedding (torch.tensor [shape=(1 + int(time // hop_length),
+                                       embedding_size)])
     """
     # Load the model if necessary
     if not hasattr(infer, 'model') or not hasattr(infer, 'capacity') or \
        (hasattr(infer, 'capacity') and infer.capacity != model):
-        infer.model = torchcrepe.Crepe(model)
-
-        # Load weights
-        file = os.path.join(
-            os.path.dirname(__file__), 'assets', f'{model}.pth')
-        infer.model.load_state_dict(torch.load(file))
-
-        infer.model.eval()
-        infer.capacity = model
+        torchcrepe.load.model(frames.device, model)
 
     # Move model to correct device (no-op if devices are the same)
     infer.model = infer.model.to(frames.device)
@@ -326,8 +465,8 @@ def postprocess(probabilities,
             Whether to also return the network confidence
 
     Returns
-        pitch (torch.tensor [shape=(1, time / hop_length)])
-        harmonicity (torch.tensor [shape=(1, time / hop_length)])
+        pitch (torch.tensor [shape=(1, 1 + int(time // hop_length))])
+        harmonicity (torch.tensor [shape=(1, 1 + int(time // hop_length))])
     """
     # Sampling is non-differentiable, so remove from graph
     probabilities = probabilities.detach()
@@ -351,17 +490,21 @@ def postprocess(probabilities,
     return pitch, harmonicity(probabilities, bins)
 
 
-def preprocess(audio, sample_rate, hop_length):
+def preprocess(audio, sample_rate, hop_length=None, batch_size=None):
     """Convert audio to model input
 
     Arguments
         audio (torch.tensor [shape=(1, time)]) - The audio signals
         sample_rate (int) - The sampling rate in Hz
         hop_length (int) - The hop_length in samples
+        batch_size (int) - The number of frames per batch
 
     Returns
-        frames (torch.tensor [shape=(time / hop_length, 1024)])
+        frames (torch.tensor [shape=(1 + int(time // hop_length), 1024)])
     """
+    # Default hop length of 10 ms
+    hop_length = sample_rate // 100 if hop_length is None else hop_length
+
     # Resample
     if sample_rate != SAMPLE_RATE:
         audio = resample(audio, sample_rate)
@@ -371,23 +514,36 @@ def preprocess(audio, sample_rate, hop_length):
     audio = torch.nn.functional.pad(audio,
                                     (WINDOW_SIZE // 2, WINDOW_SIZE // 2))
 
-    # Chunk
-    frames = torch.nn.functional.unfold(
-        audio[:, None, None, :],
-        kernel_size=(1, WINDOW_SIZE),
-        stride=(1, hop_length))
+    # Get total number of frames
+    total_frames = 1 + int(audio.size(1) // hop_length)
 
-    # shape=(batch * time / hop_length, 1024)
-    frames = frames.transpose(1, 2).reshape(-1, WINDOW_SIZE)
+    # Default to running all frames in a single batch
+    batch_size = total_frames if batch_size is None else batch_size
 
-    # Normalize
-    frames -= frames.mean(dim=1, keepdim=True)
+    # Generate batches
+    for i in range(0, total_frames, batch_size):
 
-    # Note: during silent frames, this produces very large numbers. But this
-    # seems to be what crepe expects.
-    frames /= frames.std(dim=1, keepdim=True)
+        # Convert to sample indices
+        start = i * hop_length
+        end = min(audio.size(1), (i + batch_size) * hop_length)
 
-    return frames
+        # Chunk
+        frames = torch.nn.functional.unfold(
+            audio[:, None, None, start:end],
+            kernel_size=(1, WINDOW_SIZE),
+            stride=(1, hop_length))
+
+        # shape=(batch * time / hop_length, 1024)
+        frames = frames.transpose(1, 2).reshape(-1, WINDOW_SIZE)
+
+        # Normalize
+        frames -= frames.mean(dim=1, keepdim=True)
+
+        # Note: during silent frames, this produces very large numbers. But
+        # this seems to be what crepe expects.
+        frames /= frames.std(dim=1, keepdim=True)
+
+        yield frames
 
 
 ###############################################################################
@@ -401,7 +557,7 @@ def harmonicity(probabilities, bins):
     probs_stacked = probabilities.transpose(1, 2).reshape(-1, PITCH_BINS)
 
     # shape=(batch * time / hop_length, 1)
-    bins_stacked = bins.reshape(-1, 1)
+    bins_stacked = bins.reshape(-1, 1).to(torch.int64)
 
     # Use maximum logit over pitch bins as harmonicity
     harmonicity = probs_stacked.gather(1, bins_stacked)
@@ -412,8 +568,6 @@ def harmonicity(probabilities, bins):
 
 def resample(audio, sample_rate):
     """Resample audio"""
-    import resampy
-
     # Store device for later placement
     device = audio.device
 
